# $Id: Agentalk.pm,v 1.5 1997/07/25 21:28:18 annis Exp $

=head1 NAME

Agentalk - provides an interface for sending messages to a listening
process, somewhat like C<syslog(3)>.  This module handles
communication only.  It is, by itself, of little use for anything but
the Mom system for which it was created.

=head1 VERSION

    $Revision: 1.5 $

=head1 DESCRIPTION

This is a simplified networkable interface for logging messages B<and>
data to a serving machine.  This was developed for the Mom system of
machine guarding and monitoring.

The main idea is to have a single process handling notifications by
many agent processes, daemons or fired off by cron.  "Handling" in
this case means storing, processing, and providing summaries to
processes which may request this information.

This module contains the code for both agent and monitoring processes.

=head1 Functions

B<Note:> the communication for this module is handled by Unix Domain
sockets.

=head1 C<Agentalk::agentmesg>

    $error = &Agentalk::agentmesg($agentname, $noticelevel, "$message");

The C<$agentname> identifies the agent sending the message.  The
C<$noticelevel> is one of I<note>, I<info>, I<warn>, I<error>,
I<alert>, or I<panic!> (Note the punctuation in I<panic!>).  The
C<"$message"> should B<not> be newline terminated.

C<$error> will be C<undef> if there were no problems and will contain
a string in case of errors.  That string will also be sent to
C<STDOUT> via C<warn>.

The information will come to the monitoring process in "C<::>"
separated fields, like this: C<$agentname::$noticelevel::$message>.

=head1 C<Agentalk::agentdata>

    $error = &Agentalk::agentdata($agentname, {'key', 'data'}+);

As above, C<$agentname> identifies the agent.  There must then be one
or more key/data pairs which are sent to the monitoring process.  The
error code is as in C<Agentalk::agentmesg>.

The data is sent in this fashion:

    BEGIN_DATA agent_name
    $key0::$value0
    $key1::$value1
    ...
    $keyN::$valueN
    END_DATA

B<Note> that in this case each line is newline terminated.

=head1 C<Agentalk::agentsocket>

    $error = &Agentalk::agentsocket($newsocketfile);

The default socket location is C</tmp/.agentsock>.  This will reset
that for all subsequent calls.  I wouldn't advise this unless you
really know what you're doing and why.

=head1 C<Agentalk::agentlevels>

    $error = &Agentalk::agentlevels(@new_level_list);

This resets the default levels list (see C<Agentalk::agentmesg> for a
list of defaults) to some new values.  Everytime
C<Agentalk::agentmesg> is called, the C<$level> argument is checked
against this list, raising an error if it is not in the list.  Note
that the monitoring process code does not care what the level
is... you have to write that in yourself.

=head1 C<Agentalk::listen_to_agents>

    ($error, $timeleft) = &Agentalk::listen_to_agents($subroutine);

or

    ($error, $timeleft) = &Agentalk::listen_to_agents(\&subroutine);
    ($error, $timeleft) = &Agentalk::listen_to_agents($subroutine, $timeout);

This performs the C<$subroutine> for every connection made with a
single argument, which is the filehandle to the connecting process.
It will wait C<$timeout> seconds, which may be fractional.  The form
without C<$timeout> will loop forever.

=head1 C<Agentalk::get_listener_filehandle>

    $socket_fh = &Agentalk::get_listener_filehandle();

This returns the socket filehandle for the listener.  If one is not
yet opened, one is created and that value is then returned.  This is
intended for people needing to use C<select> and who intend to write
their own connection handling loops.

=head1 C<Agentalk::datestring>

    $string = &Agentalk::datestring();
    $string = &Agentalk::datestring($some_value_from_time);

This just returns a simple "May 12 9:12:43" style date string for use
in logging.  The first form returns the string for "now."  You are, of
course, free to ignore this.

=head1 AUTHOR

William S. Annis, Biomedical Computing Group, F<annis@biostat.wisc.edu>.

=head1 BUGS and COMMENTS

This module does not handle authentication or encryption of data.  I
hope to eventually add authentication and possibly encryption if this
every uses more than Unix Domain sockets, although there is nothing to
prevent you from encrypting the messages and key/data pairs.

AS OF Wed Jul 9 12:15:25 1997, this module is not intended for use
outside the Biomedical Computing Group, Department of Biostatistics,
University of Wisconsin Medical School.  This excludes beta testers,
of course.

=cut

#----------------------------------------------------------------------
# Agentalk.pm: Communication code for logging events.
# Copyright 1997 William S. Annis
#
# Version $Revision: 1.5 $
#----------------------------------------------------------------------

package    Agentalk;
require    5.003;
require    Exporter;

@ISA     = qw(Exporter);
@EXPORT  = qw(agentmesg agentdata agentsocket agentlevels listen_to_agents get_listener_filehandle datestring);

use Socket;
use strict;

# Default Unix Domain socket.  May be overwritten with &agentsocket();
my $socket = '/tmp/.agentsock';

# Default levels.  These may also be overwritten, &agentlevels().
my @levels = qw(note info warn error alert panic!);

# This holds (a glob of) the actual server filehandle.
my $SOCKET;

# Send a message.
sub agentmesg {

    # First, make sure someone's listening.
    unless ( -S $socket or -p $socket ) {
        my $mesg = "agentmesg:'no one listening, no $socket.'  No message sent.\n";
        warn $mesg;
        return $mesg;
    }

    my($agent, $level, $message) = @_;

    my $error = &checklevel($level);
    if (defined $error) {
        warn "agentmesg:'$error', No message sent.\n";
        return $error;
    }

    if (! socket(SOCK, PF_UNIX, SOCK_STREAM, 0)) {
        warn "agentmesg:'$!', No message sent.\n";
        return "$!";
    }
    
    if (! connect(SOCK, sockaddr_un($socket)))  {
        warn "agentmesg:'$!', No message sent.\n";
        return "$!";
    }

    print SOCK "$agent\::$level\::$message";
    close SOCK;

    return undef;
}

# Send data to the server.
sub agentdata {
    # First, make sure someone's listening.
    unless ( -S $socket or -p $socket ) {
        my $mesg = "agentmesg:'no one listening, no $socket.'  No message sent.\n";
        warn $mesg;
        return $mesg;
    }

    my($agent, @data) = @_;

    if (! socket(SOCK, PF_UNIX, SOCK_STREAM, 0)) {
        warn "agentmesg:'$!', No message sent.\n";
        return "$!";
    }
    
    if (! connect(SOCK, sockaddr_un($socket)))  {
        warn "agentmesg:'$!', No message sent.\n";
        return "$!";
    }

    # Now, loop through, spewing key/data pairs.
    print SOCK "BEGIN_DATA $agent\n";
    my $i;
    for ($i = 0; $i < $#data; $i = $i + 2) {
        print SOCK "$data[$i]\::$data[$i + 1]\n";
    }
    print SOCK "END_DATA\n";
    close SOCK;

    return undef;
}

sub agentsocket {
    my $newsocket = shift;

    $socket = $newsocket;
}

# Change the level codes.  I can't see how this can cause an error in
# anything but logic, so the error code is sort of bogus.  I keep it 
# for consistency.
sub agentlevels {
    @levels = @_;
    return undef;
}

# This is the main entry for the monitor program.  It will open the 
# socket if it needs to.  Two args: $subroutine and $timeout.
# Returns (error, timeleft).
sub listen_to_agents {
    my ($subroutine, $timeout) = @_;

    # Have we opened the socket yet?
    if (! defined $SOCKET) {
        my $error = &open_listener;
        if (defined $error) {
            warn "listen_to_agents:'$error', Listen failed.\n";
            return ("listen_to_agents:'$error', Listen failed.\n", undef);
        }
    }

    # Now, what's there to listen to.
    if (defined $timeout) {
        my ($rin, $win, $ein, $nfound);
        $rin = $win = $ein = "";
        vec($rin, fileno($SOCKET), 1) = 1;
        my $timeleft;
        ($nfound, $timeleft) = select($rin, $win, $ein, $timeout);
        if ($nfound <= 0) {
            return $timeleft;
        } else {
            my $paddr = accept(Client, $SOCKET);
            &$subroutine(scalar \*Client);
            return (undef, $timeleft);
        }
    } else {
        # Loop forever.
        while (1) {
            my $paddr = accept(Client, $SOCKET);
            # If accept is interrupted under SunOS by the raising of SIGHUP,
            # you get nonsense passed to the subroutine.  The next checks for
            # that, and avoids irritating closed filehandle messages.
            next if (! defined $paddr);
            &$subroutine(scalar \*Client);
        }
    }
}

# This returns the socket that is being used to listening.  If there isn't
# one already open, one is created, and that value returned.
sub get_listener_filehandle {
    # Have we opened the socket yet?
    if (! defined $SOCKET) {
        my $error = &open_listener;
        if (defined $error) {
            warn "listen_to_agents:'$error', Listen failed.\n";
            return ("listen_to_agents:'$error', Listen failed.\n", undef);
        }
    }

    return $SOCKET;
}

# This opens the socket to listen to children.
sub open_listener {
    my $error = undef;
    my $uaddr = sockaddr_un($socket);
    my $proto = getprotobyname('tcp');

    socket(Listener, PF_UNIX, SOCK_STREAM, 0) or return "socket: $!";
    unlink $socket;
    bind(Listener, $uaddr) or return "bind: $!";
    listen(Listener, SOMAXCONN) or return "listen: $!";
    $SOCKET = scalar \*Listener;
    
    # Unadvertised, but useful for code that might fork.
    $SIG{CHLD} = \&Infanticide;

    # Without this, you get greate silliness.
    return undef;
}


# Verify that the level is in the list.  Return undef if fine, an error
# string if not.  PRIVATE function.
sub checklevel {
    my $level = shift;

    my $l;
    foreach $l (@levels) {
        if ($level eq $l) {
            return undef;
        }
    }

    return "no such level $level";
}

# Create a nicer datestring for this instant.
sub datestring {
    my $t = shift;
    my @date;

    if (defined $t) {
        @date = split / /, scalar localtime $t;
    } else {
        @date = split / /, scalar localtime time;
    }
    pop(@date);                 # Remove year.
    shift(@date);               # Remove day of the week.

    return "@date";
}

# Infanticide.  The user can redirect sigchld stuff if she wants.  Otherwise,
# this will provide zombie protection.
sub Infanticide {
    $SIG{CHLD} = \&Infanticide;
    my $waitpid = wait;
}


1;
