# $Id: Prototalk.pm,v 1.1 1998/04/30 14:31:14 annis Exp $

=head1 NAME

Prototalk - provides a simplified interface for creating TCP socket protocols.

=head1 SYNOPSIS

    use Prototalk;

=head1 VERSION

    $Revision: 1.1 $

=head1 DESCRIPTION

This is a simplified set of tools to deal with network communication.
It is not inteded for public consumption outside of the Biomedical
Computing Group.

=head1 Functions

=head1 C<Prototalk::new>

    $client = new Prototalk;
    $server = new Prototalk;

Creates a new Prototalk object, which may be either a server or a
client or both.

=head1 C<Prototalk::open>

    $client->open($hostname, $port_number);
    $client->open($hostname, $service_name);

This opens a connection to a server.  Right now, this fails hard if
the server is not willing to chat on the port specified.  If a service
name is passed as the second argument, the port number is looked up.

Returns C<undef> for no error, or for an error a string in the format:
C<function_name: Unix-errno-error>.

=head1 C<Prototalk::send>

    $client->send("Here is a message, Server.\n");

This does a buffered send (really a C<print>) to a server.

=head1 C<Prototalk::send_flushed>

    $client->send_flushed("Here is a flushed message, Server.\n");

This does an unbufferd buffered send to a server.  This should be used
when using two-way communication over a single filehandle.

=head1 C<Prototalk::close>

    $client->close()

Closes the connection to the server.

=head1 C<Prototalk::close_server>

    $server->close_server;

Closes the server socket.

=head1 C<Prototalk::readline>

    $a_line = $client->readline();

Reads a single line from the server.  Blocking.

=head1 C<Prototalk::read>

    @lines = $client->read();

Reads until EOF.

=head1 C<Prototalk::serve>

    $err = $server->serve($port_number);
    $err = $server->serve($service_name);

Do initial setup to act as a server.  Takes either a port number to
sit on, or a service (which will lookup the port number from files or
NIS).

Returns C<undef> for no error, or for an error a string in the format:
C<function_name: Unix-errno-error>.

=head1 C<Prototalk::server_loop>

    $server->server_loop(\&some_handler_function);
    $server->server_loop(\&some_handler_function, $timeout);

This is the heart of a server.  For every connection received, this
function executes C<&some_handler_function>.  If no timeout is given,
this function is never left.  It would be prudent to add a signal
handler to kill the server loop.  This currently does not fork off a
separate child process to handle the request.  I feel that option
should be left to the handler function.

C<&some_handler_function> should be a function of two arguments: the
first is the (bidirectional) filehandle just opened to the new client
and the second is the name of the client machine as returned by
C<gethostbyaddr(3n)>.  It should close the filehandle when done with
it.  Also, use C<FileHandle::flush> on the filehandle if two-way
communication is taking place over the filehandle.

The second form of this, blocks for C<$timeout> seconds or until a
connection is received, which is then handled.  The function returns
the time actually waited.  The timeout value can be in fractions of a
second (it uses C<select(2)>).

=head1 C<Prototalk::filehandle>

    $fh = $server->filehandle();

This returns the filehandle for the already opened server socket.
Useful for creating your own C<select> monstrosities.

=head1 AUTHOR

William S. Annis, Biomedical Computing Group, F<annis@biostat.wisc.edu>.

=head1 BUGS and COMMENTS

Right now, the use of C<FileHandle::flush> is the only way I know of
to prevent synchronization problems when using a single filehandle for
two-way communication.  There may be a better way.  For now, it works.

=cut

#----------------------------------------------------------------------
# Prototalk.pm: Object methods for simplified TCP communication.
# Copyright 1997 William S. Annis
# 
# Version $Revision: 1.1 $
#----------------------------------------------------------------------

package   Prototalk;
use       strict;
use       Socket;
#use       FileHandle qw(flush);
#use       POSIX qw(fflush);

# This is a royal pain in the *&%^.  The perl people can't make up
# their minds where to stick this functionality.  So I include it
# myself so this will work in multiple versions of perl.  Grumble.
# Stolen from $PERLDIST/lib/flush.pl
sub _flush {
    my($old) = select(shift);
    $| = 1;
    print "";
    $| = 0;
    select($old);
}


# Create a new prototalk object.
sub new {
    my $class = shift;
    my $self = {};
    bless $self, $class;
    
    return $self;
}

# Open a connection to the server machine. Prototalk::open('hostname', port).
# If port is text, look up the service name.
sub open {
    my ($self, $hostname, $port) = @_;

    # If port is text, lookup the service.
    if ($port !~ /\d/) {
        $port = getservbyname($port, 'tcp');
    }

    my $iaddr = inet_aton($hostname);
    if (! defined $iaddr) { return "inet_aton: $!"; }

    my $paddr = sockaddr_in($port, $iaddr);
    if (! defined $paddr) { return "sockaddr_in: $!"; }

    my $proto  = getprotobyname('tcp');
    if (! defined $proto) { return "getprotobyname: $!"; }

    socket(SOCK, PF_INET, SOCK_STREAM, $proto) or return "socket: $!";
    connect(SOCK, $paddr) or return "connect: $!";

    $self->{'SOCK'} = scalar \*SOCK;

    return undef;
}

sub send {
    my $self = shift;
    my $tosend = shift;
    my $sock = scalar $self->{'SOCK'};

    print $sock "$tosend";
}

sub send_flushed {
    my $self = shift;
    my $tosend = shift;
    my $sock = scalar $self->{'SOCK'};

    print $sock "$tosend";
    #POSIX::fflush($sock);
    &_flush($sock);
}
    
sub close {
    my $self = shift;
    my $sock = scalar $self->{'SOCK'};
    
    close $sock;
    undef $self->{'SOCK'};
}

sub close_server {
    my $self = shift;
    my $server = scalar $self->{'Server'};

    close $server;
    undef $self->{'Server'};
}

sub readline {
    my($self) = shift;
    my($sock) = $self->{'SOCK'};

    return scalar <$sock>;
}

sub read {
    my($self) = shift;
    my($sock) = $self->{'SOCK'};

    return <$sock>;
}    

sub serve {
    my($self, $port) = @_;

    if ($port !~ /\d/) {
        $port = getservbyname($port, 'tcp');
    }

    my $proto = getprotobyname('tcp');
    if (! defined $proto) { return "getprotobyname: $!"; }

    socket(Server, PF_INET, SOCK_STREAM, $proto) or return "socket: $!";
    setsockopt(Server, SOL_SOCKET, SO_REUSEADDR, pack("l", 1))
        or return "setsockopt: $!";
    bind(Server, sockaddr_in($port, INADDR_ANY)) or return "bind: $!";
    listen(Server, SOMAXCONN)                    or return "listen: $!";

    $self->{'Server'} = scalar \*Server;

    # Sets up inital reaper of dead child processes, though of course
    # The user may redefine this.  This is a but sneaky, since the
    # server loop does NOT fork by default.  If you want forking, the
    # Handler function has to do it.  This is just here for when I forget.
    $SIG{CHLD} = \&Infanticide;
    
    return undef;
}

# Goes into the everlasting loop, checking for connections, then
# performing the sub passed to it for each connection.  The routine
# Should take a single argument: the file handle for the client, which
# that subroutine should also close.  The sub may also grab the
# second argument, the name of the host connected to.
sub server_loop {
    my($self, $subroutine, $timeout) = @_;

    my $paddr;
    my $server = $self->{'Server'};
    my $loop_forever_p = 1;
    my $nfound;

    if (defined $timeout) {
        $loop_forever_p = 0;
        my ($rin, $win, $ein);
        $rin = $win = $ein = "";
        vec($rin, fileno($server), 1) = 1;
        my $timeleft;
        ($nfound, $timeleft) = select($rin, $win, $ein, $timeout);
        if ($nfound <= 0) {
            return $timeleft;
        } else {
            $paddr = accept(Client, $server);
            my ($port, $iaddr) = sockaddr_in($paddr);
            my $name = gethostbyaddr($iaddr, AF_INET);
            &$subroutine(scalar \*Client, $name);
            return ($nfound, $timeleft);
        }
    } else {
        # Loop forever on this.
        while (1) {
            $paddr = accept(Client, $server);
            my ($port, $iaddr) = sockaddr_in($paddr);
            my $name = gethostbyaddr($iaddr, AF_INET);
            &$subroutine(scalar \*Client, $name);
        }
    }
}

sub filehandle {
    my $self = shift;

    return $self->{'Server'};
}


# Infanticide.
sub Infanticide {
    $SIG{CHLD} = \&Infanticide;
    my $waitpid = wait;
}


1;
